import torch
import torch.nn as nn
import math


class MetaDropoutFast(nn.Module):
    def __init__(self, hsz):
        super().__init__()
        self.hsz = hsz

        self.Wq = nn.Linear(768, hsz)
        self.Wk = nn.Linear(768, hsz)
        self.Wl = nn.Linear(2, 10)
        self.Wf = nn.Linear(10 + 2, 1)
        self.Wi = nn.Linear(10 + 2, 1)

        self.init_weights()

    def init_weights(self):
        self.Wf.weight.data.uniform_(-0.01, 0.01)
        self.Wi.weight.data.uniform_(-0.01, 0.01)
        self.Wf.bias.data.uniform_(1, 2)
        self.Wi.bias.data.uniform_(-2, -1)
        self.Wq.bias.data.fill_(1)
        self.Wk.bias.data.fill_(1)

    def forward(self, query, key, train_eval_losses, ft_, it_, ct_):
        qt = self.Wq(query)
        kt = self.Wk(key)
        cbt = torch.matmul(qt, kt.transpose(-1, -2)) / math.sqrt(self.hsz)
        cbt = cbt.view(-1, 1)

        lt = self.Wl(train_eval_losses)
        lt = lt.expand(cbt.shape[0], -1)

        ft = torch.sigmoid(self.Wf(torch.cat([ct_, ft_, lt], dim=-1)))
        it = torch.sigmoid(self.Wi(torch.cat([ct_, it_, lt], dim=-1)))

        ct = torch.mul(ft, ct_) + torch.mul(it, cbt)

        decision_probs = torch.sigmoid(ct)
        decision_mask = gumbel_sigmoid(ct, training=True)

        return decision_mask, decision_probs, ft, it, ct


def gumbel_sigmoid(x, tau=1e-12, training=True):
    if not training:
        return (x / tau).sigmoid()

    y = x.sigmoid()

    g1 = -torch.empty_like(x).exponential_().log()
    g2 = -torch.empty_like(x).exponential_().log()
    y_hard = ((x + g1 - g2) / tau).sigmoid()

    y_hard = (y_hard - y).detach() + y
    return y_hard
